import importlib
from sachima.log import logger


class ReportsHandler(object):
    def __init__(self, **kwargs):
        if isinstance(kwargs["handler"], list):
            self.handlers = kwargs["handler"]
        else:
            self.handlers = [kwargs["handler"]]  # only one in list

    def handle(self, model_in, params):
        data_in = model_in
        for handler_str in self.handlers:
            logger.info("Call handler: " + handler_str)
            m = importlib.import_module("handler." + handler_str, handler_str)
            m = importlib.reload(m)
            # each handler should has run function
            res = m.run(data_in, params)
            logger.info(
                "HANDLER: {} RETURN TYPE: {} SIZE: {}".format(
                    handler_str, type(res), len(res)
                )
            )
            if res is not None:
                # the previous handler's result with push to next handler
                data_in = [res]
        # return the last handler's result
        return data_in



import hashlib
import operator
import pandas as pd
import datetime
from sachima.log import logger


def lengthOfLongestSubstring(s):
    """
    :type s: str
    :rtype: int
    """
    stub = -1
    max_len = 0
    cache = {}

    for i, v in enumerate(s):
        if v in cache and cache[v] > stub:
            stub = cache[v]
            cache[v] = i
        else:
            cache[v] = i
            if i - stub > max_len:
                max_len = i - stub
    return max_len


def longest_common_substring(s1, s2):
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0
    return s1[x_longest - longest : x_longest], x_longest


def extract(df, p, *cols):
    """
    用cols中的参数名从p中提取参数并过滤df
    如果参数不存在 提取下一个 直到结束
    """
    for c in cols:
        try:
            theparam = p.get(c, None)
        except KeyError:
            logger.info("params {} not exists get next...".format(c))
            theparam = None
        except:
            raise

        logger.info("extract {} by {} : {}".format(type(df), c, str(theparam)))
        if theparam == "" or theparam is None or theparam == []:
            logger.info("empty param continue...")
            continue
        # if isinstance(theparam, list):
        #     df = df[df[c].isin(theparam)]
        # if isinstance(theparam, str):
        # handle frontend param 2019-01-17T00:10:00.000Z
        if len(df) == 0:
            break
        if type(df[c].iloc[0]) is datetime.date:
            theparam = _to_date_list(theparam)
            df = df[(df[c] >= theparam[0]) & (df[c] <= theparam[1])]
        elif type(df[c].iloc[0]) is pd.Timestamp:
            if isinstance(theparam, list):
                day_begin = pd.Timestamp(theparam[0])
                day_end = pd.Timestamp(theparam[1])
                df = df[(df[c] >= day_begin) & (df[c] <= day_end)]
            else:
                day_begin = pd.Timestamp(theparam).floor(freq="D")
                day_end = pd.Timestamp(theparam).ceil(freq="D")
                df = df[(df[c] >= day_begin) & (df[c] <= day_end)]
        elif isinstance(theparam, list):
            df = df[df[c].isin(theparam)]
        else:
            df = df[df[c].isin([theparam])]
        logger.info("data remain {} lines".format(str(len(df))))
    return df


def _to_date_list(p):
    if type(p) is pd.Timestamp:
        return [p.date(), p.date()]
    elif isinstance(p, list):
        return [pd.Timestamp(val).date() for val in p]
    elif isinstance(p, str):
        return [pd.Timestamp(p).date(), pd.Timestamp(p).date()]


class Tools:
    @classmethod
    def get_md5_value(cls, s):
        myMd5 = hashlib.md5()
        myMd5.update(s.encode("utf8"))
        myMd5_Digest = myMd5.hexdigest()
        return myMd5_Digest

    @classmethod
    def special_char_remove(self, s):
        """
        特殊字符替换处理函数, 在windows下这些字符无法在文件名中存在
        """
        return (
            s.replace(",", " ")
            .replace("<", "小于")
            .replace(">", "大于")
            .replace("*", "_")
        )

    @classmethod
    def get_truth(cls, inp, relate, cut):
        ops = {
            ">": operator.gt,
            "<": operator.lt,
            ">=": operator.ge,
            "<=": operator.le,
            "=": operator.eq,
        }
        return ops[relate](inp, cut)

    @classmethod
    def maybe_float(cls, s):
        try:
            return float(s)
        except (ValueError, TypeError):
            return s

    @classmethod
    def excel_colnum_string(cls, n):
        string = ""
        while n > 0:
            n, remainder = divmod(n - 1, 26)
            string = chr(65 + remainder) + string
        return string


# if __name__ == "__main__":
# YEAR = 2017
# MONTH = 1
# DAY = 1
# HOUR = 12
# MINUTES = 12
# SECONDS = 12

# case1 = (
#     pd.Timestamp(YEAR, MONTH, DAY),
#     [datetime.date(YEAR, MONTH, DAY)],
#     "#1 single timestamp input",
# )
# case2 = (
#     pd.Timestamp(YEAR, MONTH, DAY, HOUR, MINUTES, SECONDS),
#     [datetime.date(YEAR, MONTH, DAY)],
#     "#2 single timestamp input with time",
# )
# case3 = (
#     [pd.Timestamp(YEAR, MONTH, DAY)],
#     [datetime.date(YEAR, MONTH, DAY)],
#     "#3 single timestamp list",
# )
# case4 = (
#     [pd.Timestamp(YEAR, MONTH, DAY), pd.Timestamp(YEAR, MONTH + 1, DAY)],
#     [datetime.date(YEAR, MONTH, DAY), datetime.date(YEAR, MONTH + 1, DAY)],
#     "#4 multi timestamp list",
# )
# case5 = (
#     [
#         pd.Timestamp(YEAR, MONTH, DAY),
#         "{}-{}-{}".format(str(YEAR), str(MONTH + 1), str(DAY)),
#     ],
#     [datetime.date(YEAR, MONTH, DAY), datetime.date(YEAR, MONTH + 1, DAY)],
#     "#5 mix timestamp and str list",
# )
# case6 = (
#     [
#         pd.Timestamp(YEAR, MONTH, DAY),
#         "boom!{}-{}-{}".format(str(YEAR), str(MONTH + 1), str(DAY)),
#     ],
#     [datetime.date(YEAR, MONTH, DAY), datetime.date(YEAR, MONTH + 1, DAY)],
#     ValueError,
# )

# tests = [case1, case2, case3, case4, case5, case6]

# for t in tests:
#     print(_to_date_list(t[0]))
#     assert _to_date_list(t[0]) == t[1], "{} not correct".format(t[2])

# print(longest_common_substring("曝光:不要..._新浪博客", "曝光:人|..."))



import pandas as pd
from sachima import conf
import os
import importlib
from sachima.params import set_sql_params
from sachima.log import logger
from sachima.wrappers import timer
from tqdm import tqdm
import io
import time
import logging
import sys
import threading
import itertools


@timer
def sql(sql, datatype):
    """
    return DataFrame
    """
    logger.info("running sql {}".format(sql))
    return pd.read_sql(sql, datatype)


class Data:
    def __init__(self, dataname, datatype, params, prefunc):
        """
        dataname: sql filename
        datatype: db engine  or filetype in str
        """
        self.source = dataname
        self.type = datatype
        title = dataname + " | " + str(datatype)
        logger.info("=" * 12 + " " + title + " " + "=" * 12)
        if datatype in ("xls", "xlsx"):
            self.data = pd.read_excel(
                os.path.join(conf.get("PROJ_DIR"), "data", dataname)
            )
        elif datatype in ("csv", "txt"):
            self.data = pd.read_csv(
                os.path.join(conf.get("PROJ_DIR"), "data", dataname)
            )
        elif datatype in ("api",):
            api_cls = importlib.import_module("services." + dataname, package="..")
            api = api_cls.Api()
            self.data = api.data
        else:
            # read sql file from ./sqls
            str_sql = open(
                os.path.join(conf.get("PROJ_DIR"), "sqls", dataname), encoding="utf-8",
            ).read()
            sql = str_sql
            # pre process before sql loaded
            if prefunc:
                sql = prefunc(set_sql_params(str_sql, params), params)
            else:
                sql = set_sql_params(str_sql, params)

            start = time.time()
            done = False

            def animate(log):
                elapsed_time = 0
                for c in itertools.cycle(["|", "/", "-", "\\"]):
                    if done:
                        break
                    sys.stdout.write(
                        "\r<{}> parsing sql {} {} ms".format(dataname, c, elapsed_time)
                    )
                    sys.stdout.flush()
                    time.sleep(0.1)
                    elapsed_time += 100
                log("<{}> parsing elapsed time: {} ms".format(dataname, elapsed_time))
                # sys.stdout.write("\r\n Done!     ")

            t = threading.Thread(target=animate, args=(logger.info,))
            t.daemon = True
            t.start()

            # todo: try cache if sql get error still need to run done = True
            try:
                chunks = pd.read_sql(sql, datatype, chunksize=100)
            finally:
                done = True

            t.join()
            df = pd.DataFrame()
            logger.info("<{}> start loading data... ".format(dataname))
            for chunk in tqdm(chunks, total=200):
                df = pd.concat([df, chunk])
            loading_data_elapsed_time = time.time() - start
            logger.info(
                "<{}> loading data elapsed time: {} seconds".format(
                    dataname, loading_data_elapsed_time
                )
            )
            self.data = df



from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name="sachima",
    version="2020.6.18.1",
    author="nocmk2",
    author_email="jianye.zhang@gmail.com",
    description="A package for Data Analyst",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/DessertsLab/Sachima",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "aenum",
        # "pandas",
        # "Flask",
        # "nameko",
        # "redis",
        # "cython >= 0.29.6",
        # "thriftpy >= 0.3.9",
        # "impyla == 0.16.2",
        # "pymysql",
        # "SQLAlchemy",
        "tqdm",
    ],
    include_package_data=True,
    package_data={"": ["*.css"]},
)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sachima.proto

from google.protobuf import symbol_database as _symbol_database
from google.protobuf import reflection as _reflection
from google.protobuf import message as _message
from google.protobuf import descriptor as _descriptor
import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="sachima.proto",
    package="sachima",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\rsachima.proto\x12\x07sachima"\x1f\n\rReportRequest\x12\x0e\n\x06params\x18\x01 \x01(\t"\x1e\n\x0bReportReply\x12\x0f\n\x07message\x18\x01 \x01(\t2E\n\x08Reporter\x12\x39\n\tRunReport\x12\x16.sachima.ReportRequest\x1a\x14.sachima.ReportReplyb\x06proto3'
    ),
)


_REPORTREQUEST = _descriptor.Descriptor(
    name="ReportRequest",
    full_name="sachima.ReportRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="params",
            full_name="sachima.ReportRequest.params",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=26,
    serialized_end=57,
)


_REPORTREPLY = _descriptor.Descriptor(
    name="ReportReply",
    full_name="sachima.ReportReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message",
            full_name="sachima.ReportReply.message",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=59,
    serialized_end=89,
)

DESCRIPTOR.message_types_by_name["ReportRequest"] = _REPORTREQUEST
DESCRIPTOR.message_types_by_name["ReportReply"] = _REPORTREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ReportRequest = _reflection.GeneratedProtocolMessageType(
    "ReportRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_REPORTREQUEST,
        __module__="sachima_pb2"
        # @@protoc_insertion_point(class_scope:sachima.ReportRequest)
    ),
)
_sym_db.RegisterMessage(ReportRequest)

ReportReply = _reflection.GeneratedProtocolMessageType(
    "ReportReply",
    (_message.Message,),
    dict(
        DESCRIPTOR=_REPORTREPLY,
        __module__="sachima_pb2"
        # @@protoc_insertion_point(class_scope:sachima.ReportReply)
    ),
)
_sym_db.RegisterMessage(ReportReply)


_REPORTER = _descriptor.ServiceDescriptor(
    name="Reporter",
    full_name="sachima.Reporter",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    serialized_start=91,
    serialized_end=160,
    methods=[
        _descriptor.MethodDescriptor(
            name="RunReport",
            full_name="sachima.Reporter.RunReport",
            index=0,
            containing_service=None,
            input_type=_REPORTREQUEST,
            output_type=_REPORTREPLY,
            serialized_options=None,
        )
    ],
)
_sym_db.RegisterServiceDescriptor(_REPORTER)

DESCRIPTOR.services_by_name["Reporter"] = _REPORTER

# @@protoc_insertion_point(module_scope)



import os
import functools
import time
import sachima.sns as sns
import datetime
from sachima.log import logger
from sachima import conf
import inspect


def only_in_night(func):
    def wrapper(a):
        if 7 <= datetime.now().hour < 22:
            pass
        else:
            func(a)

    return wrapper


def send(func):
    """
    send msg to sns app called dingding
    https://open-doc.dingtalk.com/docs/doc.htm?spm=a219a.7629140.0.0.2vzwCr&treeId=257&articleId=105735&docType=1
    you should conif your token in sachima.yaml
    """
    ERROR_GRP_TOKEN = conf.get("SNS_DINGDING_ERROR_GRP_TOKEN")
    INFO_GRP_TOKEN = conf.get("SNS_DINGDING_INFO_GRP_TOKEN")
    SENDING_STR = conf.get("SNS_DINGDING_SENDING_STR")
    ERRSENT_STR = conf.get("SNS_DINGDING_ERRSENT_STR")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        file_name = os.path.basename(inspect.getsourcefile(func))
        time_str = str(datetime.datetime.now())
        try:
            t = SENDING_STR.format(file_name, time_str)
            sns.send_dingding(t, t, INFO_GRP_TOKEN)
            value = func(*args, **kwargs)
            return value
        except Exception as ex:
            title = ERRSENT_STR.format(file_name, time_str)
            exception_info = str(ex)
            sns.send_dingding(title, title + exception_info, ERROR_GRP_TOKEN)
            sns.send_dingding(title, title + exception_info, INFO_GRP_TOKEN)

    return wrapper


def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()  # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3
        # logger.info(f"Finished {inspect.getfile(func)} in {run_time:.4f} secs")
        logger.info(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


def debug(func):
    """Print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]  # 1
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]  # 2
        signature = ", ".join(args_repr + kwargs_repr)  # 3
        logger.info(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        logger.info(f"{func.__name__!r} returned {value!r}")  # 4
        return value

    return wrapper_debug


def slow_down(func):
    """Sleep 1 second before calling the function"""

    @functools.wraps(func)
    def wrapper_slow_down(*args, **kwargs):
        time.sleep(1)
        return func(*args, **kwargs)

    return wrapper_slow_down


def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


def clock(func):
    @functools.wraps(func)
    def wrapper_clock(*args, **kwargs):
        t0 = time.time()
        result = func(*args, **kwargs)
        elapsed = time.time() - t0
        name = func.__name__
        arg_lst = []
        if args:
            arg_lst.append(", ".join(repr(arg) for arg in args))
        if kwargs:
            pairs = ["%s=%r" % (k, w) for k, w in sorted(kwargs.items())]
            arg_lst.append(", ".join(pairs))
        arg_str = ", ".join(arg_lst)
        logger.info("[%0.8fs] %s(%s) => %r" % (elapsed, name, arg_str, result))
        return result

    return wrapper_clock
